"""Module for convenience functions that go beyond the basic backend capabilities."""

import posixpath
from copy import deepcopy
import t2kdm
from six import print_
import sys, sh
from contextlib import contextmanager
import re

def strip_output(output):
    """Turn the output of a command into an iterable.

    Returns one element per line, stripped of leading and trailing white
    spaces. Requires an iterable command output.
    """

    for line in output:
        line = line.strip()
        if len(line) > 0:
            yield line

class Replica(object):
    """Simple helper class that contains information of a replica."""
    def __init__(self, line):
        """Initialise from a line of output of `replicas -l`."""
        x = line.split()
        if len(x) > 1:
            # The full information
            self.SE_name = x[0]
            self.SE_type = x[1]
            self.status = x[2]
            self.checksum = x[3]
            self.path = x[4]
        else:
            # Just the replica url
            self.path = x[0]
            SE = t2kdm.storage.get_SE(x[0])
            if SE is not None:
                self.SE_name = SE.name
                self.SE_type = SE.type
            else:
                self.SE_name = 'UNKOWN'
                self.SE_type = '?'
            self.status = '?'
            self.checksum = '-' # Not a '?', this would trigger a checksum error

class ReplicaState(object):
    """Deal with the state of a file's replicas."""

    def __init__(self, remotepath, full=True):
        """Get the ReplicaState of a given remotepath.

        If `full` is `False`, only check what replicas exist, but no additional information.
        """

        self.replicas = []
        for line in strip_output(t2kdm.replicas(remotepath, long=(not full), _iter=True)):
            self.replicas.append(Replica(line))

    def get_file_categories(self):
        """Return a list of categories the file falls into.

        Supported categories are:

            all - all files are in this category
            checksum_error - do the checksums of the replicas differ?
            no_replica - There is no replica of this file at all
            SE_<SE_name> - all files replicated on SE <SE_name>

        """

        # List of the categories
        cat = []

        # Set of checksums
        checksums = set()

        # All files are in 'all'
        cat.append('all')

        # Add all SEs
        for rep in self.replicas:
            checksums.add(rep.checksum)
            cat.append('SE_%s'%(rep.SE_name))

        c = len(checksums)
        if c == 0:
            cat.append('no_replica')
        if c > 1 or '?' in checksums or u'?' in checksums:
            cat.append('checksum_error')

        return cat

class CheckResult(object):
    """Class to contain the results of a replica check.

    It contains sets of all checked files in a directory. Each set
    corresponds to a certain category, e.g. all checked files, files
    with checksum problems, files replicated to a specific SE, etc.

    It can also contain further CheckResults of subdirectories.
    """

    def __init__(self, path):
        """Initialise the CheckResult.

        It will have to be filled with actual results afterwards.
        """

        self.path = path

        # Dictionary of sets
        self.file_sets = {}

        # List of results of subdirectories
        self.subresults = []

    def add_file_to_set(self, f, s, dictionary=None):
        """Add a file to the specified set.

        If the set does not exist yet, create it.
        """

        if dictionary is None:
            d = self.file_sets

        if s not in d:
            self.file_sets[s] = set()

        d[s].add(f)

    def add(self, other):
        """Add files from other CheckResult to this one."""

        # Add files
        for s in other.file_sets:
            for f in other.file_sets[s]:
                self.add_file_to_set(f, s)

        # Add sub results
        self.subresults.extend(other.subresults)

    def add_subresult(self, check_result):
        """Add a CheckResult of a sub directory."""
        self.subresults.append(check_result)

    def collect_file_sets(self):
        """Return file sets with all files including subdirectories."""

        # Copy own sets so we do not modify them
        ret = deepcopy(self.file_sets)

        # Add all sets from sub-results
        for subresult in self.subresults:
            # Get collected sub-results
            col = subresult.collect_file_sets()
            # Go through all sets and add to return value
            for s in col:
                if s not in ret:
                    ret[s] = set()

                # Update the results
                ret[s] |= col[s]

        return ret

    @staticmethod
    def _report_checksum(sets, list_faulty=True):
        """Report faulty checksums."""
        try:
            faulty = sets['checksum_error']
        except KeyError:
            faulty = set()
        n_err = len(faulty)
        report = "%d files have faulty checksums.\n"%(n_err,)
        if list_faulty:
            for f in sorted(list(faulty)):
                report += " -> %s\n"%(f,)
        return report

    @staticmethod
    def _report_SE(sets, SE, list_faulty=True):
        """Report replication on given SE."""

        if SE not in t2kdm.storage.SE_by_name and SE != 'UNKNOWN':
            raise ValueError("Not a valid SE: %s"%(SE,))

        try:
            all = (sets['all'])
        except KeyError:
            all = set()
        try:
            replicated = sets['SE_'+SE]
        except KeyError:
            replicated = set()
        faulty = all - replicated
        n_rep = len(replicated)
        n_err = len(faulty)
        report = "%d files have been replicated on %s.\n"%(n_rep, SE)
        report += "%d files are missing from %s.\n"%(n_err, SE)
        if list_faulty:
            for f in sorted(list(faulty)):
                report += " -> %s\n"%(f,)
        return report

    def report(self, *args, **kwargs):
        """Create a report of the result.

        The arguments determine what is reported on and in what order.
        Valid report types:

            checksum - report files with checksum errors
            SE_<SE_name> - report files that are *not* replicated on the given SE

        If no report types are provided, the following reports are generated:

            checksum

        If the keyword argument `list_faulty` is `True` (default), all
        problematic files are listed. Otherwise only the number of faulty files
        is reported.
        """

        # Get total sets of all files in all directories
        sets = self.collect_file_sets()

        # Use standard list of reports if none are provided
        if len(args) == 0:
            args = ['checksum']
        list_faulty = kwargs.pop('list_faulty', True)

        # Print total number of files
        try:
            n_all = len(sets['all'])
        except KeyError:
            n_all = 0
        report = "Checked a total of %d files in %s.\n"%(n_all, self.path)

        # Add requested reports
        for arg in args:
            if arg == 'checksum':
                report += self._report_checksum(sets, list_faulty=list_faulty)
            elif arg.startswith('SE_'):
                report += self._report_SE(sets, arg[3:], list_faulty=list_faulty)
            else:
                raise ValueError("Unknown report type.""")

        return report

    def list_faulty(self, *args):
        """Return a list of faulty files, according to the provided checks (see `report`)."""
        faulty = set()
        for line in self.report(*args).split('\n'):
            if line.startswith(' -> '):
                faulty.add(line[4:])
        return sorted(list(faulty))

def check_replicas(remotepath, recursive=False, progress=False, quick=False):
    """Check the replicas of a given path.

    Returns a CheckResult.

    Arguments
    ---------

    recursive: bool/string. All files and sub-directories of a given path are replicated.
               If `recursive` is a string, it is treated as regular expression
               and only matching subfolders or files are replicated.
               Default: `False`
    progress: bool. Print out progress information on screen.
              Default: `False`
    quick: bool. Do not do the checksum test. Makes things go faster.
           Default `False`
    """

    # Need to differentiate between progress in base calla nd recursive calls
    if progress is True: # Base call, set progress to 2
        progress = 2
    elif progress == 2: # Recursive call, set progress to 1
        progress = 1
    elif progress is False: # No progress requested, set to 0
        progress = 0

    result = CheckResult(remotepath)

    # Do thing recursively if requested
    if isinstance(recursive, str):
        regex = re.compile(recursive)
        recur = True
    else:
        regex = None
        recur = recursive

    if t2kdm.backend.is_dir(remotepath):
        # Go through the contents of the directory
        if progress == 2:
            print_("Checking files. This might take a while...")
        for f in strip_output(t2kdm.ls(remotepath, _iter=True)):
            if regex is not None and not regex.search(f):
                continue # Skip files/directories that do not match the regex
            newpath = posixpath.join(remotepath, f)
            if t2kdm.backend.is_dir(newpath):
                # This is a directory
                if recur:
                    # Add subresult of checking subdirectory recursively
                    result.add_subresult(check_replicas(newpath, recursive=recursive, progress=progress))
            else:
                # Check file replicas and add to current result
                if progress > 0:
                    print_('.', end='')
                    sys.stdout.flush() # Make sure every dot is printed directly to screen
                result.add(check_replicas(newpath))
        if progress == 2:
            print_("\nDone.")
    else:
        # Handle single file
        rep_state = ReplicaState(remotepath, full=not quick)

        for category in rep_state.get_file_categories():
            result.add_file_to_set(remotepath, category)

    return result

def check(remotepath, checksum=False, se=[],
        recursive=False, nolist=False, list=None, **kwargs):
    """Check a remote path for problems with the replicas.

    Arguments
    ---------

    checksum: bool. Check whether the checksums of all replicas are identical.
              Default `False`
    se: list. List of SE names for which a replication report should be printed.
        Default []
    recursive: bool. Go through sub directories recursively.
               Default `False`
    nolist: bool. Do not list the faulty files in te report.
            Default `False`
    list: string. Store a list of problematic files in the given filename.
          Default `None`

    Behaves like (kinda) like an `sh` command.
    """

    tests = []
    if checksum:
        tests.append('checksum')
    for s in se:
        SE = t2kdm.storage.get_SE(s)
        if SE is None:
            raise sh.ErrorReturnCode_1('', '',
                    "Not a valid storage element: %s\n"%(s,))
        tests.append('SE_'+SE.name)

    if len(tests) == 0:
        raise sh.ErrorReturnCode_1('', '',
                "Nothing to report! You need to specify at least on check to be performed.\n"%(s,))

    # Test the files and get the result
    result = check_replicas(remotepath, recursive=recursive, progress=True, quick=not checksum)

    if list is not None:
        # Save the list of faulty files
        with open(list, 'wt') as f:
            for fault in result.list_faulty(*tests):
                f.write('%s\n'%(fault,))

    report = result.report(*tests, list_faulty=not nolist)
    it = kwargs.pop('_iter', False)
    if not it:
        # Just return the check report
        return report
    else:
        # We need to provide a generator that yields the lines of the putput
        return (line+'\n' for line in report.split('\n')[:-1]) # Last element of split is always empty string



"""T2K Data Manager

Helpful tools to manage the T2K data on the grid.
"""

import configuration
import backends
import storage
import utils
import sys

if sys.argv[0].endswith('t2kdm-config'):
    # Someone is calling t2kdm-config.
    # Do not try to load any configuration, as that is what they might be trying to fix!
    pass
else:
    # Load configuration
    config = configuration.load_config()
    # Get backend according to configuration
    backend = backends.get_backend(config)

    # Get functions from backend
    ls = backend.ls
    replicas = backend.replicas
    replicate = backend.replicate
    remove = backend.remove
    get = backend.get
    put = backend.put



import sh
import itertools
import posixpath
import os
from t2kdm import storage
import re

class CommandChain(object):
    """Class that executes function calls in sequence.

    It expects the functions to return iterables and will return an iterable when called.
    """

    def __init__(self):
        """Initialise CommandChain.

        Actual commands have to be added with `add`.
        """

        self.commands = []
        self.args = []
        self.kwargs = []

    def add(self, cmd, *args, **kwargs):
        """Add a command o the chain."""

        self.commands.append(cmd)
        self.args.append(args)
        self.kwargs.append(kwargs)

    def __call__(self):
        """Yield from all commands in turn."""
        for c, a, k in zip(self.commands, self.args, self.kwargs):
            for line in c(*a, **k):
                yield line

class GridBackend(object):
    """Class that handles the actual work on the grid.

    This is just a base class that other classes must inherit from.

    The convention for remote paths is that public methods expect "t2k paths",
    i.e. paths within the t2k grid directory, omitting the common prefix.
    Internal methods, i.e. the ones beginning with an underscore '_',
    expect the full path to be passed to them.
    """

    def __init__(self, **kwargs):
        """Initialise backend.

        Accepts the follwoing keyword arguments:

        basedir: String. Default: '/t2k.org'
            Sets the base directory of the backend.
            All paths are specified relative to that position.
        """

        self.basedir = kwargs.pop('basedir', '/t2k.org')
        if len(kwargs) > 0:
            raise TypeError("Invalid keyword arguments: %s"%(list(kwargs.keys),))

    def should_raise(self, code=1, **kwargs):
        """Analyse kwargs and decide whether an exception should be raised in case of an error."""

        ok = kwargs.pop('_ok_code', [])
        if code in ok:
            return False
        else:
            return True

    def error(self, message, **kwargs):
        """Handle errors by either raising an Exception or just printing the error message."""

        if self.should_raise(code=1, **kwargs):
            raise sh.ErrorReturnCode_1('', '', message)
        else:
            return self._iterable_output_from_text(message, **kwargs)

    def full_path(self, relpath):
        """Prepend the base dir to a path."""
        return posixpath.normpath(self.basedir + relpath)

    def _ls(self, remotepath, **kwargs):
        raise NotImplementedError()

    def ls(self, remotepath, **kwargs):
        """List contents of a remote logical path.

        Supported keyword arguments:

        long: Bool. Default: False
            Print a longer, more detailed listing.
        directory: Bool. Default: False
            List directory entries instead of contents.
        """
        _path = self.full_path(remotepath)
        return self._ls(_path, **kwargs)

    def _is_dir(self, remotepath):
        """Is the remote path a directory?"""
        return str(self._ls(remotepath, long=True, directory=True)).strip()[0] == 'd'

    def is_dir(self, remotepath):
        """Is the remote path a directory?"""
        return self._is_dir(self.full_path(remotepath))

    def _replica_state(self, storagepath, **kwargs):
        """Internal method to get the state of a replica, e.g. 'ONLINE'."""
        raise NotImplementedError()

    def _replica_checksum(self, storagepath, **kwargs):
        """Internal method to get the checksum of a replica."""
        raise NotImplementedError()

    def _replicas(self, remotepath, **kwargs):
        raise NotImplementedError()

    def _add_replica_info(self, rep):
        SE = storage.get_SE_by_path(rep)
        if SE is not None:
            return "%-32s %-4s %-8s %8.8s %s"%(SE.name, SE.type, self._replica_state(rep), self._replica_checksum(rep), rep)
        else:
            return "%-32s %-4s %-8s %8.8s %s"%('UNKNOWN', '?', self._replica_state(rep), self._replica_checksum(rep), rep)

    @staticmethod
    def _iterable_output_from_text(text, **kwargs):
        """Turn a block of text into an iterable if necessary."""
        it = kwargs.pop('_iter', False)
        if it == False:
            # Nothing to do here
            return text
        else:
            # Split by lines and return iterable
            lines = text.split('\n')
            if text[-1] == '\n':
                # Remove last empty string from split
                lines = lines[:-1]
            return (line+'\n' for line in lines)

    @staticmethod
    def _iterable_output_from_iterable(iterable, **kwargs):
        """Turn an iterable into a block of text if necessary."""
        it = kwargs.pop('_iter', False)
        if it == True:
            # Nothing to do here
            return iterable
        else:
            # Concatenate all lines
            text = ""
            for line in iterable:
                text += line
            return text

    def replicas(self, remotepath, **kwargs):
        """List replicas of a remote logical path.

        Supported keyword arguments:

        long: Bool. Default: False
            Print a longer, more detailed listing.
        """
        _path = self.full_path(remotepath)
        l = kwargs.pop('long', False)
        if l:
            # Parse each line and add additional information
            it = kwargs.pop('_iter', False)
            kwargs['_iter'] = True
            return self._iterable_output_from_iterable(
                    (self._add_replica_info(line) for line in self._replicas(_path, **kwargs)),
                    _iter=it)
        else:
            return self._replicas(_path, **kwargs)

    def _bringonline(self, storagepath, timeout, **kwargs):
        raise NotImplementedError()

    def bringonline(self, replica, **kwargs):
        """Make sure the given replica is online"""

        # We need to do multiple tries with short timeouts,
        # because it seems like the bringonline commands do not notice when they succeed.
        timeout = 10
        tries = 1440
        for i in range(tries):
            try:
                self._bringonline(replica, timeout)
            except sh.ErrorReturnCode:
                # Did not work
                # Try again
                continue
            except sh.SignalException_SIGSEGV:
                # Did not work
                # Try again
                continue
            else:
                # It worked
                # exit loop
                break
        else:
            # Never reached a break
            # It did not work so we raise an error
            return self.error("Could not bring replica online.\n", **kwargs)

        return self._iterable_output_from_text("Replica is online.\n", **kwargs)

    def _replicate(self, source_storagepath, destination_storagepath, **kwargs):
        raise NotImplementedError()

    def replicate(self, remotepath, destination, source=None, tape=False, recursive=False, **kwargs):
        """Replicate the file to the specified storage element.

        If no source storage elment is provided, the closest replica is chosen.
        If `tape` is `True`, tape SEs are considered when choosing the closest one.
        If `recursive` is `True`, all files and sub-directories of a given path are replicated.
        if `recursive` is a string, it is treated as regular expression
        and only matching subfolders or files are replicated.
        """

        # Do thing recursively if requested
        if isinstance(recursive, str):
            regex = re.compile(recursive)
            recursive = True
        else:
            regex = None
        _path = self.full_path(remotepath)
        if recursive and self._is_dir(_path):
            # Go through the contents of the directory recursively
            it = kwargs.pop('_iter', False)
            newpaths = []
            for element in self.ls(remotepath, _iter=True):
                element = element.strip()
                if regex is None or regex.search(element):
                    newpaths.append(posixpath.join(remotepath, element))
            def outputs(paths):
                for path in paths:
                    # Ignore errors and segfaults when running recursively
                    ok = kwargs.pop('_ok_code', None)
                    ex = kwargs.pop('_bg_exc', None)
                    try:
                        yield self.replicate(path, destination, source, tape, recursive,
                                _iter=True, _ok_code=list(range(-255,256)), _bg_exc=False, **kwargs)
                    except sh.ErrorReturnCode:
                        pass
                    except sh.SignalException_SIGSEGV:
                        pass
                    if ok is not None:
                        kwargs['_ok_code'] = ok
                    if ex is not None:
                        kwargs['_bg_exc'] = ex
            iterable = itertools.chain.from_iterable(outputs(newpaths))
            return self._iterable_output_from_iterable(iterable, _iter=it)

        # Get destination SE and check if file is already present
        dst = storage.get_SE(destination)
        if dst is None:
            return self.error("Could not find storage element %s.\n", **kwargs)

        if dst.has_replica(remotepath):
            # Replica already at destination, nothing to do here
            return self._iterable_output_from_text(
                    "%s\nReplica already present at destination storage element %s.\n"%(remotepath, dst.name,), **kwargs)

        # Get source SE
        if source is None:
            src = dst.get_closest_SE(remotepath, tape=tape)
            if src is None:
                return self.error("Could not find valid storage element with replica of %s.\n"%(remotepath,), **kwargs)
        else:
            src = storage.get_SE(source)
            if src is None:
                return self.error("Could not find storage element %s.\n"%(source,), **kwargs)

            if not src.has_replica(remotepath):
                # Replica not present at source, throw error
                return self.error("%s\nNo replica present at source storage element %s\n"%(remotepath, src.name,), **kwargs)

        source_path = src.get_replica(remotepath)
        destination_path = dst.get_storage_path(remotepath)

        # Chain commands for bringing the file online and replicating it
        it = kwargs.pop('_iter', False)
        kwargs['_iter'] = True # Need to iterate to make command chaining possible
        chain = CommandChain()
        if src.type == 'tape':
            chain.add(self._iterable_output_from_text, "Bringing online %s\n"%(source_path,), **kwargs)
            chain.add(self.bringonline, source_path, **kwargs)
        chain.add(self._replicate, source_path, destination_path, **kwargs)
        return self._iterable_output_from_iterable(chain(), _iter=it)

    def _get(self, storagepath, localpath, **kwargs):
        raise NotImplementedError()

    def get(self, remotepath, localpath, source=None, tape=False, recursive=False, force=False, **kwargs):
        """Download a file from the grid.

        If no source storage elment is provided, the closest replica is chosen.
        If `tape` is True, tape SEs are considered when choosing the closest one.
        If `recursive` is `True`, all files and sub-directories of a given path are replicated.
        If `recursive` is a string, it is treated as regular expression
        and only matching subfolders or files are replicated.
        If `force` is `True`, local files will be overwritten.
        """

        # Do thing recursively if requested
        if isinstance(recursive, str):
            regex = re.compile(recursive)
            recursive = True
        else:
            regex = None
        _path = self.full_path(remotepath)
        if recursive and self._is_dir(_path):
            # Go through the contents of the directory recursively
            it = kwargs.pop('_iter', False)
            newpaths = []
            for element in self.ls(remotepath, _iter=True):
                element = element.strip()
                if regex is None or regex.search(element):
                    _remote = posixpath.join(remotepath, element)
                    _local = os.path.join(localpath, element)
                    newpaths.append( (_remote, _local) )
            def outputs(paths):
                for rpath, lpath in paths:
                    # Ignore errors and segfaults when running recursively
                    ok = kwargs.pop('_ok_code', None)
                    ex = kwargs.pop('_bg_exc', None)
                    try:
                        yield self.get(rpath, lpath, source, tape, recursive, force,
                                _iter=True, _ok_code=list(range(-255,256)), _bg_exc=False, **kwargs)
                    except sh.ErrorReturnCode:
                        pass
                    except sh.SignalException_SIGSEGV:
                        pass
                    if ok is not None:
                        kwargs['_ok_code'] = ok
                    if ex is not None:
                        kwargs['_bg_exc'] = ex
            iterable = itertools.chain.from_iterable(outputs(newpaths))
            return self._iterable_output_from_iterable(iterable, _iter=it)

        # Do not overwrite files unless explicitly told to
        if os.path.isfile(localpath) and not force:
            if recursive == False:
                # Raise for a single file
                return self.error("File does already exist: %s.\n"%(localpath,))
            else:
                # Print status for recursive
                return self._iterable_output_from_text(
                        "File does already exist: %s.\n"%(localpath,), **kwargs)

        if source is None:
            # Get closest SE
            SE = storage.get_closest_SE(remotepath, tape=tape)
            if SE is None:
                return self.error("Could not find valid storage element with replica of %s.\n"%(remotepath,), **kwargs)
        else:
            # Use the provided source
            SE = storage.get_SE(source)
            if SE is None:
                return self.error("Could not find storage element %s.\n"%(source,), **kwargs)

            if not SE.has_replica(remotepath):
                # Replica not present at source, throw error
                return self.error("%s\nNo replica present at source storage element %s\n"%(remotepath, SE.name,), **kwargs)

        # Get the source replica
        replica = SE.get_replica(remotepath)

        # Append the basename to the localpath if it is a directory
        if os.path.isdir(localpath):
            localpath = os.path.join(localpath, posixpath.basename(remotepath))

        # Chain commands for bringing the file online and replicating it
        it = kwargs.pop('_iter', False)
        kwargs['_iter'] = True # Need to iterate to make command chaining possible
        chain = CommandChain()
        if SE.type == 'tape':
            chain.add(self._iterable_output_from_text, "Bringing online %s\n"%(replica,), **kwargs)
            chain.add(self.bringonline, replica, **kwargs)
        chain.add(self._get, replica, localpath, **kwargs)
        return self._iterable_output_from_iterable(chain(), _iter=it)

    def _put(self, localpath, remotepath, storagepath, **kwargs):
        raise NotImplementedError()

    def put(self, localpath, remotepath, destination=None, tape=False, **kwargs):
        """Upload and register a file.

        If no destination storage element is provided, the closest one will be chosen.
        """

        # Split the local path in dir and file
        path, base = posixpath.split(localpath)

        # If the remotepath ends with '/', append the filename to it
        if remotepath.endswith('/'):
            remotepath += base

        # Get the destination
        if destination is None:
            # Get closest SE
            SE = storage.get_closest_SE(tape=tape)
            if SE is None:
                return self.error("Could not find valid storage element\n", **kwargs)
        else:
            # Use the provided destination
            SE = storage.get_SE(destination)
            if SE is None:
                return self.error("Could not find storage element %s.\n"%(destination,), **kwargs)

        # Get the storage path
        surl = SE.get_storage_path(remotepath)

        # Upload and register the file
        _path = self.full_path(remotepath)
        return self._put(localpath, _path, surl, **kwargs)

    def _remove(self, storagepath, **kwargs):
        raise NotImplementedError()

    def remove(self, remotepath, destination, recursive=False, final=False, **kwargs):
        """Remove the replica of a file from a storage element.

        If `recursive` is `True`, all files and sub-directories of a given path are removed.
        if `recursive` is a string, it is treated as regular expression
        and only matching subfolders or files are replicated.

        This command will refuse to remove the last replica of a file
        unless the `final` argument is `True`!
        """

        # Do thing recursively if requested
        if isinstance(recursive, str):
            regex = re.compile(recursive)
            recursive = True
        else:
            regex = None
        _path = self.full_path(remotepath)
        if recursive and self._is_dir(_path):
            # Go through the contents of the directory recursively
            it = kwargs.pop('_iter', False)
            newpaths = []
            for element in self.ls(remotepath, _iter=True):
                element = element.strip()
                if regex is None or regex.search(element):
                    newpaths.append(posixpath.join(remotepath, element))
            def outputs(paths):
                for path in paths:
                    # Ignore errors and segfaults when running recursively
                    ok = kwargs.pop('_ok_code', None)
                    ex = kwargs.pop('_bg_exc', None)
                    try:
                        yield self.remove(path, destination, recursive=recursive,
                                _iter=True, _ok_code=list(range(-255,256)), _bg_exc=False, **kwargs)
                    except sh.ErrorReturnCode:
                        pass
                    except sh.SignalException_SIGSEGV:
                        pass
                    if ok is not None:
                        kwargs['_ok_code'] = ok
                    if ex is not None:
                        kwargs['_bg_exc'] = ex
            iterable = itertools.chain.from_iterable(outputs(newpaths))
            return self._iterable_output_from_iterable(iterable, _iter=it)

        # Get destination SE and check if file is already not present
        dst = storage.get_SE(destination)
        if dst is None:
            return self.error("Could not find storage element %s.\n"%(destination,), **kwargs)

        if not dst.has_replica(remotepath):
            # Replica already not present at destination, nothing to do here
            return self._iterable_output_from_text(
                    "%s\nReplica not present at destination storage element %s.\n"%(remotepath, dst.name,), **kwargs)

        # Check how many replicas there are
        # If it is only one, refuse to delete it
        nrep = self.replicas(remotepath).count('\n') # count lines
        if not final and nrep <= 1:
            return self.error("Only one replica of file left! Aborting.\n", **kwargs)

        destination_path = dst.get_replica(remotepath)

        return self._remove(destination_path, **kwargs)

class LCGBackend(GridBackend):
    """Grid backend using the LCG command line tools `lfc-*` and `lcg-*`."""

    def __init__(self, **kwargs):
        # LFC paths alway put a '/grid' as highest level directory.
        # Let us not expose that to the user.
        kwargs['basedir'] = '/grid'+kwargs.pop('basedir', '/t2k.org')
        GridBackend.__init__(self, **kwargs)

        self._proxy_init_cmd = sh.Command('voms-proxy-init')
        self._ls_cmd = sh.Command('lfc-ls')
        self._replicas_cmd = sh.Command('lcg-lr')
        self._replica_state_cmd = sh.Command('lcg-ls')
        self._replica_checksum_cmd = sh.Command('lcg-get-checksum')
        self._bringonline_cmd = sh.Command('lcg-bringonline')
        self._replicate_cmd = sh.Command('lcg-rep')
        self._cp_cmd = sh.Command('lcg-cp')
        self._cr_cmd = sh.Command('lcg-cr')
        self._del_cmd = sh.Command('lcg-del')

    def _ls(self, remotepath, **kwargs):
        # Translate keyword arguments
        l = kwargs.pop('long', False)
        d = kwargs.pop('directory', False)
        args = []
        if l:
            args.append('-l')
        if -d:
            args.append('-d')
        args.append(remotepath)

        return self._ls_cmd(*args, **kwargs)

    def _replicas(self, remotepath, **kwargs):
        return(self._replicas_cmd('lfn:'+remotepath, **kwargs))

    def _replica_state(self, storagepath, **kwargs):
        _path = storagepath.strip()
        it = kwargs.pop('_iter', None)
        try:
            listing = self._replica_state_cmd('-l', _path, **kwargs)
        except sh.ErrorReturnCode:
            listing = '- - - - - ?'
        except sh.SignalException_SIGSEGV:
            listing = '- - - - - ?'
        return listing.split()[5]

    def _replica_checksum(self, storagepath, **kwargs):
        _path = storagepath.strip()
        it = kwargs.pop('_iter', None)
        try:
            listing = self._replica_checksum_cmd(_path, **kwargs)
        except sh.ErrorReturnCode:
            listing = '? -'
        except sh.SignalException_SIGSEGV:
            listing = '? -'
        try:
            checksum = listing.split()[0]
        except IndexError:
            # Something weird happened
            checksum = '?'
        return checksum

    @staticmethod
    def _ignore_identical_lines(iterable, **kwargs):
        last_line = None
        for line in iterable:
            if line == last_line:
                continue
            else:
                last_line = line
                yield line

    def _bringonline(self, storagepath, timeout, **kwargs):
        kwargs['_err_to_out'] = True # Verbose output is on stderr
        kwargs.pop('_err', None) # Cannot specify _err and _err_ro_out at same time

        # Get original command output
        return self._bringonline_cmd('-v', '--bdii-timeout', timeout, '--srm-timeout', timeout, '--sendreceive-timeout', timeout, '--connect-timeout', timeout, storagepath, **kwargs)

    def _replicate(self, source_storagepath, destination_storagepath, **kwargs):
        kwargs['_err_to_out'] = True # Verbose output is on stderr
        kwargs.pop('_err', None) # Cannot specify _err and _err_ro_out at same time
        it = kwargs.pop('_iter', False) # should the out[put be an iterable?
        kwargs['_iter'] = True # Need iterable to ignore identical lines

        # Get original command output
        iterable = self._replicate_cmd('-v', '--sendreceive-timeout', 14400, '--checksum', '-d', destination_storagepath, source_storagepath, **kwargs)

        # Ignore lines that are identical to the previous
        iterable = self._ignore_identical_lines(iterable)

        # return requested kind of output
        return self._iterable_output_from_iterable(iterable, _iter=it)

    def _get(self, storagepath, localpath, **kwargs):
        kwargs['_err_to_out'] = True # Verbose output is on stderr
        kwargs.pop('_err', None) # Cannot specify _err and _err_ro_out at same time
        it = kwargs.pop('_iter', False) # should the out[put be an iterable?
        kwargs['_iter'] = True # Need iterable to ignore identical lines

        # Get original command output
        iterable = self._cp_cmd('-v', '--sendreceive-timeout', 14400, '--checksum', storagepath, localpath, **kwargs)

        # Ignore lines that are identical to the previous
        iterable = self._ignore_identical_lines(iterable)

        # return requested kind of output
        return self._iterable_output_from_iterable(iterable, _iter=it)

    def _put(self, localpath, remotepath, storagepath, **kwargs):
        kwargs['_err_to_out'] = True # Verbose output is on stderr
        kwargs.pop('_err', None) # Cannot specify _err and _err_ro_out at same time
        it = kwargs.pop('_iter', False) # should the out[put be an iterable?
        kwargs['_iter'] = True # Need iterable to ignore identical lines

        # Get original command output
        iterable = self._cr_cmd('-v', '--sendreceive-timeout', 14400, '--checksum', '-d', storagepath, '-l', remotepath, localpath, **kwargs)

        # Ignore lines that are identical to the previous
        iterable = self._ignore_identical_lines(iterable)

        # return requested kind of output
        return self._iterable_output_from_iterable(iterable, _iter=it)

    def _remove(self, storagepath, **kwargs):
        kwargs['_err_to_out'] = True # Verbose output is on stderr
        kwargs.pop('_err', None) # Cannot specify _err and _err_ro_out at same time
        it = kwargs.pop('_iter', False) # should the out[put be an iterable?
        kwargs['_iter'] = True # Need iterable to ignore identical lines

        # Get original command output
        iterable = self._del_cmd('-v', storagepath, **kwargs)

        # Ignore lines that are identical to the previous
        iterable = self._ignore_identical_lines(iterable)

        # return requested kind of output
        return self._iterable_output_from_iterable(iterable, _iter=it)

def get_backend(config):
    """Return the backend according to the provided configuration."""

    if config.backend == 'lcg':
        return LCGBackend(basedir = config.basedir)
    else:
        raise config.ConfigError('backend', "Unknown backend!")



"""Tests for the T2K data Manager."""

import argparse
from six import print_
import t2kdm
import t2kdm.commands as cmd
import t2kdm.cli
import t2kdm.storage
from contextlib import contextmanager
import sys, os, sh
import tempfile
import posixpath

testdir = '/test/t2kdm'
testfiles = ['test1.txt', 'test2.txt']
testSEs = ['UKI-SOUTHGRID-RALPP-disk', 'UKI-SOUTHGRID-OX-HEP-disk', 'RAL-LCG22-tape']

@contextmanager
def no_output(redirect=True):
    stdout = sys.stdout
    stderr = sys.stderr
    with open('/dev/null', 'w') as null:
        try:
            if redirect:
                sys.stdout = null
                sys.stderr = null
            yield
        finally:
            sys.stdout = stdout
            sys.stderr = stderr

@contextmanager
def fake_argv(fake_argv):
    true_argv = sys.argv
    try:
        sys.argv = fake_argv
        yield
    finally:
        sys.sargv = true_argv

@contextmanager
def temp_dir():
    tempdir = tempfile.mkdtemp()
    try:
        yield tempdir
    finally:
        sh.rm('-r', tempdir)

def run_read_only_tests(backend = t2kdm.backend):
    print_("Testing ls...")
    assert(testfiles[0] in backend.ls(testdir))
    assert(testfiles[0] in backend.ls(testdir, long=True))

    print_("Testing replicas...")
    # Test short output
    assert('srm-t2k.gridpp.rl.ac.uk' in backend.replicas('/nd280/raw/ND280/ND280/00000000_00000999/nd280_00000001_0000.daq.mid.gz'))
    # Test long output
    assert('RAL-LCG22-tape' in backend.replicas('/nd280/raw/ND280/ND280/00000000_00000999/nd280_00000001_0000.daq.mid.gz', long=True))

    print_("Testing StorageElement...")
    # Test distance calculation
    assert(t2kdm.storage.SEs[0].get_distance(t2kdm.storage.SEs[1]) < 0)
    # Test getting SE by host
    assert('srm-t2k.gridpp.rl.ac.uk' in t2kdm.storage.SE_by_host['srm-t2k.gridpp.rl.ac.uk'].get_storage_path('/nd280/test'))
    # Test getting the closest SE
    t2kdm.storage.get_closest_SE('/nd280/raw/ND280/ND280/00000000_00000999/nd280_00000001_0000.daq.mid.gz')

    print_("Testing TriumfStorageElement...")
    # Test special case of TRIUMF SE
    assert('t2ksrm.nd280.org/nd280data/' in t2kdm.storage.SE_by_host['t2ksrm.nd280.org'].get_storage_path('/nd280/test'))

    print_("Testing get...")
    with temp_dir() as tempdir:
        path = posixpath.join(testdir, testfiles[0])
        filename = os.path.join(tempdir, testfiles[0])

        # Test choosing source SE automatically
        t2kdm.get(path, tempdir)
        assert(os.path.isfile(filename))
        os.remove(filename)

        # Test providing the source SE (RAL tape!)
        t2kdm.get(path, tempdir, source=testSEs[2])
        assert(os.path.isfile(filename))
        os.remove(filename)

        # Test recursive get
        t2kdm.get(testdir, tempdir, recursive=True)
        assert(os.path.isfile(filename))

    print_("Testing check...")
    with temp_dir() as tempdir:
        filename = os.path.join(tempdir, 'faulty.txt')
        with no_output():
            t2kdm.utils.check(testdir, checksum=True, se=testSEs, recursive=True, nolist=True, list=filename)
        assert os.path.isfile(filename)

    print_("Testing Commands...")
    with open('/dev/null', 'w') as out:
        cmd.ls.run_from_cli('-l /', _out=out) # This should work, but return `False`
        cmd.ls.run_from_cli('.', _out=out) # This should work

    with no_output(True):
        cmd.ls.run_from_cli('abc') # This should not work, but not throw exception
        cmd.ls.run_from_cli('"abc') # This should not work, but not throw exception
        with fake_argv(['t2kdm-ls']):
            assert(cmd.ls.run_from_console() == 0) # This should work
        with fake_argv(['t2kdm-cli']):
            assert(cmd.ls.run_from_console() == 0) # This should work
        with fake_argv(['t2kdm-ls', '/abcxyz']):
            assert(cmd.ls.run_from_console() == 1) # This should not work, hence the 1 return value

        # None of the Commands should return True in the CLI
        for com in cmd.all_commands:
            assert(com.run_from_cli('') == False)

    print_("Testing CLI...")
    cli = t2kdm.cli.T2KDmCli()
    with no_output(True):
        cli.onecmd('help ls')
        cli.onecmd('ls .')
        cli.onecmd('cd /nd280')
        cli.onecmd('cd ..')
        cli.onecmd('cd /abcxyz')
        cli.onecmd('lcd /')
        cli.onecmd('lcd .')
        cli.onecmd('lcd /root')
        cli.onecmd('lcd /abcxyz')
        cli.onecmd('lls .')
        cli.onecmd('lls ".')
        assert(cli.completedefault('28', 'ls nd28', 0, 0) == ['280'])
        assert(cli.completedefault('s', 'lls us', 0, 0) == ['sr'])
        assert(cli.completedefault('"us', 'lls "us', 0, 0) == [])

def run_read_write_tests(backend = t2kdm.backend):
    print_("Testing replicate...")
    t2kdm.replicate(testdir, testSEs[1], recursive=r'^test[1]\.t.t$')
    t2kdm.replicate(testdir, testSEs[1], recursive=r'^test[2]\.t.t$', source=testSEs[2])

    print_("Testing put...")
    with temp_dir() as tempdir:
        tempf = 'thisfileshouldnotbehereforlong.txt'
        filename = os.path.join(tempdir, tempf)
        remotename = posixpath.join(testdir, tempf)
        # Make sure the file does not exist
        try:
            for SE in t2kdm.storage.SEs:
                t2kdm.remove(remotename, SE.name, final=True)
        except sh.ErrorReturnCode:
            pass
        # Prepare something to upload
        with open(filename, 'wt') as f:
            f.write("This is testfile #3.\n")
        with no_output():
            t2kdm.put(filename, testdir+'/', destination=testSEs[0])

    print_("Testing disk SEs...")
    # Replicate test file to all SEs, to see if they all work
    for SE in t2kdm.storage.SEs:
        if SE.type == 'tape' or 'TRIUMF' in SE.name or 'KEK' in SE.name or 'QMUL' in SE.name:
            # These SEs do not seem to cooperate
            continue
        t2kdm.replicate(remotename, SE.name)

    print_("Testing remove...")
    t2kdm.remove(testdir, testSEs[1], recursive=True) # Remove everything from SE1
    # Remove uploaded file from previous test
    try:
        # This should fail!
        for SE in t2kdm.storage.SEs:
            t2kdm.remove(remotename, SE.name)
    except sh.ErrorReturnCode_1:
        pass
    else:
        raise Exception("The last copy should not have been removed!")
    # With the `final` argument it should work
    try:
        for SE in t2kdm.storage.SEs:
            t2kdm.remove(remotename, SE.name, final=True)
    except sh.ErrorReturnCode_1 as e:
        # The command will fail when the file no longer exists
        assert("No such file or directory" in e.stderr)

def run_tests():
    """Test the functions of the t2kdm."""

    parser = argparse.ArgumentParser(description="Run tests for the T2K Data Manager.")
    parser.add_argument('-w', '--write', action='store_true',
        help="do write tests. Default: read only")

    args = parser.parse_args()
    run_read_only_tests()
    if args.write:
        run_read_write_tests()

    print_("All done.")

if __name__ == '__main__':
    run_tests()



"""Module to organise storage elements."""

import posixpath
import t2kdm

class StorageElement(object):
    """Representation of a grid storage element"""

    def __init__(self, name, host, type, location, basepath):
        """Initialise StorageElement.

        `name`: Identifier for element
        `host`: Hostname of element
        `type`: Storage type of element ('tape' or 'disk')
        `location`: Location of the SE, e.g. '/europe/uk/ral'
        `basepath`: Base path for standard storage paths on element
        """

        self.name = name
        self.host = host
        self.basepath = basepath
        self.location = location
        self.type = type

    def get_storage_path(self, remotepath):
        """Generate the standard storage path for this SE from a logical file name."""
        if remotepath[0] != '/':
            raise ValueError("Remote path needs to be absolute, not relative!")
        return (self.basepath + remotepath).strip()

    def get_distance(self, other):
        """Return the distance to another StorageElement.

        Returns a negative number. The smaller (i.e. more negative) it is,
        the closer the two SE are together.
        """

        common = posixpath.commonprefix([self.location.lower()+'/', other.location.lower()+'/'])
        # The more '/' are in the common prefix, the closer the SEs are.
        # So we can take the negative number as measure of distance.
        distance = -common.count('/')
        return distance

    def get_replica(self, remotepath):
        """Return the replica of the file on this SM."""
        for rep in t2kdm.replicas(remotepath, _iter=True):
            if self.host in rep:
                return rep.strip()
        # Replica not found
        return None

    def has_replica(self, remotepath):
        """Check whether the remote path is replicated on this SE."""
        return self.host in t2kdm.replicas(remotepath)

    def get_closest_SE(self, remotepath=None, tape=False):
        """Get the storage element with the closest replica.

        If `tape` is False (default), prefer disk SEs over tape SEs.
        If no `rempotepath` is provided, just return the closest SE over all.
        """
        closest_SE = None
        closest_distance = None

        if remotepath is None:
            candidates = SEs
        else:
            candidates = []
            for rep in t2kdm.replicas(remotepath, _iter=True):
                candidates.append(get_SE_by_path(rep))

        for SE in candidates:
            if SE is None:
                continue
            if closest_SE is None:
                # Always accept the first SE
                closest_SE = SE
                closest_distance = self.get_distance(SE)
            elif SE.type != 'tape':
                # We always accept non-tape SEs
                if self.get_distance(SE) <= closest_distance:
                    # Also when they are equally close as the current choice
                    closest_SE = SE
                    closest_distance = self.get_distance(SE)
            elif tape or closest_SE.type == 'tape':
                # We accept the candidate tape SEs
                # Either bu choice or because the current best candidate is already a tape SE
                if self.get_distance(SE) < closest_distance:
                    # But only if it is *better* than the current choice
                    closest_SE = SE
                    closest_distance = self.get_distance(SE)

        return closest_SE

    def __str__(self):
        return "%s (%s) [%s]"%(self.name, self.host, self.location)

class TriumfStorageElement(StorageElement):
    """Special case of StorageElement for TRIUMF.

    Storage file paths do not translate one-to-one to logical file paths,
    so we have to catch these differences.
    """

    def get_storage_path(self, remotepath):
        """Generate the standard storage path for this SE from a logical file name."""

        # ND280 data is in the sub folder `nd280data`
        if remotepath.startswith('/nd280/'):
            return StorageElement.get_storage_path(self, '/nd280data/' + remotepath[7:])

        #Everything else seems to be one-to-one
        return StorageElement.get_storage_path(self, remotepath)

# Add actual SEs
SEs = [
    StorageElement('RAL-LCG22-tape',
        host = 'srm-t2k.gridpp.rl.ac.uk',
        type = 'tape',
        location = '/europe/uk/ral',
        basepath = 'srm://srm-t2k.gridpp.rl.ac.uk/castor/ads.rl.ac.uk/prod/t2k.org'),
    StorageElement('UKI-SOUTHGRID-RALPP-disk',
        host = 'heplnx204.pp.rl.ac.uk',
        type = 'disk',
        location = '/europe/uk/ral',
        basepath = 'srm://heplnx204.pp.rl.ac.uk/pnfs/pp.rl.ac.uk/data/t2k/t2k.org'),
    StorageElement('UKI-SOUTHGRID-OX-HEP-disk',
        host = 't2se01.physics.ox.ac.uk',
        type = 'disk',
        location = '/europe/uk/ox',
        basepath = 'srm://t2se01.physics.ox.ac.uk/dpm/physics.ox.ac.uk/home/t2k.org'),
    StorageElement('UKI-NORTHGRID-SHEF-HEP-disk',
        host = 'lcgse0.shef.ac.uk',
        type = 'disk',
        location = '/europe/uk/shef',
        basepath = 'srm://lcgse0.shef.ac.uk/dpm/shef.ac.uk/home/t2k.org'),
    StorageElement('UKI-NORTHGRID-LANCS-HEP-disk',
        host = 'fal-pygrid-30.lancs.ac.uk',
        type = 'disk',
        location = '/europe/uk/lancs',
        basepath = 'srm://fal-pygrid-30.lancs.ac.uk/dpm/lancs.ac.uk/home/t2k.org'),
    StorageElement('UKI-NORTHGRID-MAN-HEP-disk',
        host = 'bohr3226.tier2.hep.manchester.ac.uk',
        type = 'disk',
        location = '/europe/uk/man',
        basepath = 'srm://bohr3226.tier2.hep.manchester.ac.uk/dpm/tier2.hep.manchester.ac.uk/home/t2k.org'),
    StorageElement('UKI-NORTHGRID-LIV-HEP-disk',
        host = 'hepgrid11.ph.liv.ac.uk',
        type = 'disk',
        location = '/europe/uk/liv',
        basepath = 'srm://hepgrid11.ph.liv.ac.uk/dpm/ph.liv.ac.uk/home/t2k.org'),
    StorageElement('UKI-LT2-IC-HEP-disk',
        host = 'gfe02.grid.hep.ph.ic.ac.uk',
        type = 'disk',
        location = '/europe/uk/london/ic',
        basepath = 'srm://gfe02.grid.hep.ph.ic.ac.uk/pnfs/hep.ph.ic.ac.uk/data/t2k'),
    StorageElement('UKI-LT2-QMUL2-disk',
        host = 'se03.esc.qmul.ac.uk',
        type = 'disk',
        location = '/europe/uk/london/qmul',
        basepath = 'srm://se03.esc.qmul.ac.uk/t2k.org'),
    StorageElement('IN2P3-CC-disk',
        host = 'polgrid4.in2p3.fr',
        type = 'disk',
        location = '/europe/fr/in2p3',
        basepath = 'srm://polgrid4.in2p3.fr/dpm/in2p3.fr/home/t2k.org'),
    StorageElement('PIC-disk',
        host = 'srm.pic.es',
        type = 'disk',
        location = '/europe/es/pic',
        basepath = 'srm://srm.pic.es/pnfs/pic.es/data/t2k.org'),
    #StorageElement('IFIC-LCG2-disk',
    #    host = 'srmv2.ific.uv.es',
    #    type = 'disk',
    #    location = '/europe/es/ific',
    #    basepath = ''), # Need to know base path
    #StorageElement('INFN-BARI1-disk',
    #    host = 'storm-se-01.ba.infn.it',
    #    type = 'disk',
    #    location = '/europe/',
    #    basepath = ''), # Need to know base path
    TriumfStorageElement('CA-TRIUMF-T2K1-disk',
        host = 't2ksrm.nd280.org',
        type = 'disk',
        location = '/americas/ca/triumf',
        basepath = 'srm://t2ksrm.nd280.org'),
    StorageElement('JP-KEK-CRC-02-disk',
        host = 'kek2-se01.cc.kek.jp',
        type = 'disk',
        location = '/asia/jp/kek',
        basepath = 'srm://kek2-se01.cc.kek.jp/t2k.org'),
# TODO (Possibly Dirac SEs that are not available to T2K in general):
#GRIF-disk
#GridPPSandboxSE
#Nebraska1-disk
#UNIBE-LHEP-disk
    ]

SE_by_name = {}
SE_by_host = {}

for SE in SEs:
    SE_by_name[SE.name] = SE
    SE_by_host[SE.host] = SE

def get_SE_by_path(path):
    """Return the StorageElement corresponsing to the given srm-path."""
    for SE in SEs:
        if SE.host in path:
            return SE
    return None

def get_SE(SE):
    """Get the StorageElement by all means necessary."""
    if SE in SE_by_name:
        return SE_by_name[SE]
    if SE in SE_by_host:
        return SE_by_host[SE]
    return get_SE_by_path(SE)

def get_closest_SE(remotepath=None, location=None, tape=False):
    """Get the closest storage element with a replica of the given file.

    If `tape` is False (default), prefer disk SEs over tape SEs.
    If no `rempotepath` is provided, just return the closest SE over all.
    """

    if location is None:
        location = t2kdm.config.location

    # Create a psude SE with the correct location
    SE = StorageElement('local',
        host = 'localhost',
        type = 'disk',
        location = location,
        basepath = '/')

    return SE.get_closest_SE(remotepath, tape=tape)

def list_storage_elements(**kwargs):
    """Print all available storage elments on screen.

    Behaves (kinda) like an `sh` command.
    """

    ret = ( str(SE)+'\n' for SE in SEs )
    it = kwargs.pop('_iter', False)
    if not it:
        # Glue single lines together
        ''.join(ret)

    return ret
